/*
 * Copyright 2006 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.mcarthur.sandy.commons.pool.composite;

import org.apache.commons.pool.ObjectPool;

import java.io.ObjectStreamException;
import java.io.Serializable;
import java.util.NoSuchElementException;

/**
 * Specifies the behavor of the pool when the pool is out of idle objects.
 *
 * @see CompositeObjectPoolFactory#setExhaustionBehavior(ExhaustionBehavior)
 * @see CompositeKeyedObjectPoolFactory#setExhaustionBehavior(ExhaustionBehavior)
 * @author Sandy McArthur
 * @since #.#
 * @version $Revision$ $Date$
 */
public final class ExhaustionBehavior implements Serializable {

    private static final long serialVersionUID = -4895490364329810018L;

    /**
     * Grow the pool when all idle objects have been exhausted.
     */
    public static final ExhaustionBehavior GROW = new ExhaustionBehavior("GROW");

    /**
     * Throw a {@link NoSuchElementException} when all idle objects have been exhaused. Clients of the poll must
     * call {@link ObjectPool#addObject()} to prefill the pool.
     */
    public static final ExhaustionBehavior FAIL = new ExhaustionBehavior("FAIL");

    /**
     * enum name.
     */
    private final String name;

    private ExhaustionBehavior(final String name) {
        this.name = name;
    }

    public String toString() {
        return name;
    }

    // Autogenerated with Java 1.5 enums
    public static ExhaustionBehavior[] values() {
        return new ExhaustionBehavior[] {FAIL, GROW};
    }

    // necessary for serialization
    private static int nextOrdinal = 0;
    private final int ordinal = nextOrdinal++;
    private Object readResolve() throws ObjectStreamException {
        return values()[ordinal];
    }
}